# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from enum import Enum
from dataclasses import dataclass
from typing import List, Optional, ClassVar, Union

from . import values


AstType = Union['Node', values.ValueType, list]
ScalarAstType = Union['Node', int, float]
SpatialAstType = Union['Node', values.SpatialValueType]
TemporalAstType = Union['Node', values.TemporalValueType]
ArrayAstType = Union['Node', List[AstType]]


class Node:
    """ The base class for all other nodes to display the AST of CQL.
    """
    inline: bool = False

    def get_sub_nodes(self) -> List[AstType]:
        """ Get a list of sub-node of this node.

            :return: a list of all sub-nodes
            :rtype: list[Node]
        """
        return []

    def get_template(self) -> str:
        """ Get a template string (using the ``.format`` method)
            to represent the current node and sub-nodes. The template string
            must provide a template replacement for each sub-node reported by
            :func:`~pygeofilter.ast.Node.get_sub_nodes`.

            :return: the template to render
        """
        raise NotImplementedError

    def __eq__(self, other) -> bool:
        if type(self) != type(other):
            return False

        self_dict = {
            k: v.__geo_interface__ if hasattr(v, '__geo_interface__') else v
            for k, v in self.__dict__.items()
        }
        other_dict = {
            k: v.__geo_interface__ if hasattr(v, '__geo_interface__') else v
            for k, v in other.__dict__.items()
        }
        return self_dict == other_dict


class Condition(Node):
    """ The base class for all nodes representing a condition
    """
    pass


class Not(Condition):
    """
    Node class to represent a negation condition.

    :ivar sub_node: the condition node to be negated
    :type sub_node: Node
    """

    def __init__(self, sub_node: Node):
        self.sub_node = sub_node

    def get_sub_nodes(self) -> List[AstType]:
        """ Returns the sub-node for the negated condition. """
        return [self.sub_node]

    def get_template(self) -> str:
        return "NOT {}"


class CombinationOp(Enum):
    AND = "AND"
    OR = "OR"


@dataclass
class Combination(Condition):
    """ Node class to represent a condition to combine two other conditions
        using either AND or OR.
    """

    lhs: Node
    rhs: Node
    op: ClassVar[CombinationOp]

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs, self.rhs]

    def get_template(self) -> str:
        return f"{{}} {self.op.name} {{}}"

    @classmethod
    def from_items(cls, first, *rest) -> Node:
        result = first
        for item in rest:
            result = cls(result, item)
        return result


@dataclass
class And(Combination):
    op: ClassVar[CombinationOp] = CombinationOp.AND


@dataclass
class Or(Combination):
    op: ClassVar[CombinationOp] = CombinationOp.OR


class Predicate(Node):
    """ The base class for all nodes representing a predicate
    """
    pass


class ComparisonOp(Enum):
    EQ = '='
    NE = '<>'
    LT = '<'
    LE = '<='
    GT = '>'
    GE = '>='


@dataclass
class Comparison(Predicate):
    """ Node class to represent a comparison predicate: to compare two
        expressions using a comparison operation.
    """

    lhs: ScalarAstType
    rhs: ScalarAstType
    op: ClassVar[ComparisonOp]

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs, self.rhs]

    def get_template(self) -> str:
        return f"{{}} {self.op.value} {{}}"


@dataclass
class Equal(Comparison):
    op: ClassVar[ComparisonOp] = ComparisonOp.EQ


@dataclass
class NotEqual(Comparison):
    op: ClassVar[ComparisonOp] = ComparisonOp.NE


@dataclass
class LessThan(Comparison):
    op: ClassVar[ComparisonOp] = ComparisonOp.LT


@dataclass
class LessEqual(Comparison):
    op: ClassVar[ComparisonOp] = ComparisonOp.LE


@dataclass
class GreaterThan(Comparison):
    op: ClassVar[ComparisonOp] = ComparisonOp.GT


@dataclass
class GreaterEqual(Comparison):
    op: ClassVar[ComparisonOp] = ComparisonOp.GE


@dataclass
class Between(Predicate):
    """ Node class to represent a BETWEEN predicate: to check whether an
        expression value within a range.
    """

    lhs: Node
    low: ScalarAstType
    high: ScalarAstType
    not_: bool

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs, self.low, self.high]

    def get_template(self) -> str:
        return f"{{}} {'NOT ' if self.not_ else ''}BETWEEN {{}} AND {{}}"


@dataclass
class Like(Predicate):
    """ Node class to represent a wildcard sting matching predicate.
    """

    lhs: Node
    pattern: str
    nocase: bool
    wildcard: str
    singlechar: str
    escapechar: str
    not_: bool

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs]

    def get_template(self) -> str:
        return (
            f"{{}} {'NOT ' if self.not_ else ''}"
            f"{'I' if self.nocase else ''}LIKE '{self.pattern}'"
            # TODO wildcard, singlechar, escapechar
        )


@dataclass
class In(Predicate):
    """ Node class to represent list checking predicate.
    """
    lhs: AstType
    sub_nodes: List[AstType]
    not_: bool

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs] + list(self.sub_nodes)

    def get_template(self) -> str:
        return (
            f"{{}} {'NOT ' if self.not_ else ''}IN "
            f"{', '.join(['{}'] * len(self.sub_nodes))}"
        )


@dataclass
class IsNull(Predicate):
    """ Node class to represent null check predicate.
    """

    lhs: AstType
    not_: bool

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs]

    def get_template(self) -> str:
        return f"{{}} IS {('NOT ' if self.not_ else '')}NULL"


@dataclass
class Exists(Predicate):
    lhs: AstType
    not_: bool

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs]

    def get_template(self) -> str:
        return f"{{}} {('DOES-NOT-EXIST' if self.not_ else 'EXISTS')}"


@dataclass
class Include(Predicate):
    not_: bool

    def get_template(self) -> str:
        return 'EXCLUDE' if self.not_ else 'INCLUDE'


# http://docs.opengeospatial.org/DRAFTS/19-079.html#enhanced-temporal-operators

# BEFORE                <======>     <----->    AFTER
# MEETS                         <---------->    METBY
# TOVERLAPS                 <-------------->    OVERLAPPEDBY
# BEGINS                <------------------>    BEGUNBY
# DURING            <---------------------->    TCONTAINS
# TENDS             <---------->                ENDEDBY
# TEQUALS               <------>                TEQUALS
# DISJOINT: If a proper interval T1 is intervalDisjoint another proper
#   interval T2,then the beginning of T1 is after the end of T2, or the end of
#   T1 is before the beginning of T2, i.e. the intervals do not overlap in any
#   way, but their ordering relationship is not known.

# https://github.com/geotools/geotools/blob/main/modules/library/cql/ECQL.md#temporal-predicate
# BEFORE_OR_DURING  <----->
# DURING_OR_AFTER           <----->

class TemporalComparisonOp(Enum):
    DISJOINT = 'DISJOINT'
    AFTER = 'AFTER'
    BEFORE = 'BEFORE'
    BEGINS = 'BEGINS'
    BEGUNBY = 'BEGUNBY'
    TCONTAINS = 'TCONTAINS'
    DURING = 'DURING'
    ENDEDBY = 'ENDEDBY'
    ENDS = 'ENDS'
    TEQUALS = 'TEQUALS'
    MEETS = 'MEETS'
    METBY = 'METBY'
    TOVERLAPS = 'TOVERLAPS'
    OVERLAPPEDBY = 'OVERLAPPEDBY'

    BEFORE_OR_DURING = 'BEFORE OR DURING'
    DURING_OR_AFTER = 'DURING OR AFTER'


@dataclass
class TemporalPredicate(Predicate):
    """ Node class to represent temporal predicate.
    """

    lhs: TemporalAstType
    rhs: TemporalAstType
    op: ClassVar[TemporalComparisonOp]

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs, self.rhs]

    def get_template(self) -> str:
        return f"{{}} {self.op} {{}}"


@dataclass
class TimeAfter(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.AFTER


@dataclass
class TimeBefore(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.BEFORE


@dataclass
class TimeBegins(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.BEGINS


@dataclass
class TimeBegunBy(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.BEGUNBY


@dataclass
class TimeContains(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.TCONTAINS


@dataclass
class TimeDuring(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.DURING


@dataclass
class TimeEndedBy(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.ENDEDBY


@dataclass
class TimeEnds(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.ENDS


@dataclass
class TimeEquals(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.TEQUALS


@dataclass
class TimeMeets(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.MEETS


@dataclass
class TimeMetBy(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.METBY


@dataclass
class TimeOverlaps(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.TOVERLAPS


@dataclass
class TimeOverlappedBy(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.OVERLAPPEDBY


@dataclass
class TimeBeforeOrDuring(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.BEFORE_OR_DURING


@dataclass
class TimeDuringOrAfter(TemporalPredicate):
    op: ClassVar[TemporalComparisonOp] = TemporalComparisonOp.DURING_OR_AFTER


class ArrayComparisonOp(Enum):
    AEQUALS = 'AEQUALS'
    ACONTAINS = 'ACONTAINS'
    ACONTAINEDBY = 'ACONTAINEDBY'
    AOVERLAPS = 'AOVERLAPS'


@dataclass
class ArrayPredicate(Predicate):
    """ Node class to represent array predicates.
    """

    lhs: ArrayAstType
    rhs: ArrayAstType
    op: ClassVar[ArrayComparisonOp]

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs, self.rhs]

    def get_template(self) -> str:
        return f"{{}} {self.op} {{}}"


@dataclass
class ArrayEquals(ArrayPredicate):
    op: ClassVar[ArrayComparisonOp] = ArrayComparisonOp.AEQUALS


@dataclass
class ArrayContains(ArrayPredicate):
    op: ClassVar[ArrayComparisonOp] = ArrayComparisonOp.ACONTAINS


@dataclass
class ArrayContainedBy(ArrayPredicate):
    op: ClassVar[ArrayComparisonOp] = ArrayComparisonOp.ACONTAINEDBY


@dataclass
class ArrayOverlaps(ArrayPredicate):
    op: ClassVar[ArrayComparisonOp] = ArrayComparisonOp.AOVERLAPS


class SpatialComparisonOp(Enum):
    INTERSECTS = 'INTERSECTS'
    DISJOINT = 'DISJOINT'
    CONTAINS = 'CONTAINS'
    WITHIN = 'WITHIN'
    TOUCHES = 'TOUCHES'
    CROSSES = 'CROSSES'
    OVERLAPS = 'OVERLAPS'
    EQUALS = 'EQUALS'


@dataclass
class SpatialComparisonPredicate(Predicate):
    """ Node class to represent spatial relation predicate.
    """

    lhs: SpatialAstType
    rhs: SpatialAstType
    op: ClassVar[SpatialComparisonOp]

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs, self.rhs]

    def get_template(self) -> str:
        return f"{self.op.name}({{}}, {{}})"


@dataclass
class GeometryIntersects(SpatialComparisonPredicate):
    op: ClassVar[SpatialComparisonOp] = SpatialComparisonOp.INTERSECTS


@dataclass
class GeometryDisjoint(SpatialComparisonPredicate):
    op: ClassVar[SpatialComparisonOp] = SpatialComparisonOp.DISJOINT


@dataclass
class GeometryContains(SpatialComparisonPredicate):
    op: ClassVar[SpatialComparisonOp] = SpatialComparisonOp.CONTAINS


@dataclass
class GeometryWithin(SpatialComparisonPredicate):
    op: ClassVar[SpatialComparisonOp] = SpatialComparisonOp.WITHIN


@dataclass
class GeometryTouches(SpatialComparisonPredicate):
    op: ClassVar[SpatialComparisonOp] = SpatialComparisonOp.TOUCHES


@dataclass
class GeometryCrosses(SpatialComparisonPredicate):
    op: ClassVar[SpatialComparisonOp] = SpatialComparisonOp.CROSSES


@dataclass
class GeometryOverlaps(SpatialComparisonPredicate):
    op: ClassVar[SpatialComparisonOp] = SpatialComparisonOp.OVERLAPS


@dataclass
class GeometryEquals(SpatialComparisonPredicate):
    op: ClassVar[SpatialComparisonOp] = SpatialComparisonOp.EQUALS


@dataclass
class Relate(Predicate):
    """ Node class to represent spatial relation predicate.
    """

    lhs: SpatialAstType
    rhs: SpatialAstType
    pattern: str

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs, self.rhs]

    def get_template(self) -> str:
        return f"RELATE({{}}, {{}}, '{self.pattern}')"


class SpatialDistanceOp(Enum):
    DWITHIN = 'DWITHIN'
    BEYOND = 'BEYOND'


@dataclass
class SpatialDistancePredicate(Predicate):
    """ Node class to represent spatial relation predicate.
    """

    lhs: SpatialAstType
    rhs: SpatialAstType
    distance: float
    units: str
    op: ClassVar[SpatialDistanceOp]

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs, self.rhs]

    def get_template(self) -> str:
        return f"{self.op.name}({{}}, {{}}, {self.distance}, '{self.units}')"


@dataclass
class DistanceWithin(SpatialDistancePredicate):
    op: ClassVar[SpatialDistanceOp] = SpatialDistanceOp.DWITHIN


@dataclass
class DistanceBeyond(SpatialDistancePredicate):
    op: ClassVar[SpatialDistanceOp] = SpatialDistanceOp.BEYOND


@dataclass
class BBox(Predicate):
    """ Node class to represent a bounding box predicate.
    """

    lhs: SpatialAstType
    minx: float
    miny: float
    maxx: float
    maxy: float
    crs: Optional[str] = None

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs]

    def get_template(self) -> str:
        return (
            f"BBOX({{}}, {self.minx}, {self.miny}, {self.maxx}, "
            f"{self.maxy}, {repr(self.crs)})"
        )


class Expression(Node):
    """ The base class for all nodes representing expressions
    """
    pass


class Attribute(Expression):
    """ Node class to represent attribute lookup expressions

        :ivar name: the name of the attribute to be accessed
        :type name: str
    """
    inline = True

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f"ATTRIBUTE {self.name}"


class ArithmeticOp(Enum):
    ADD = '+'
    SUB = '-'
    MUL = '*'
    DIV = '/'


@dataclass
class Arithmetic(Expression):
    """ Node class to represent arithmetic operation expressions with two
        sub-expressions and an operator.
    """

    lhs: ScalarAstType
    rhs: ScalarAstType
    op: ClassVar[ArithmeticOp]

    def get_sub_nodes(self) -> List[AstType]:
        return [self.lhs, self.rhs]

    def get_template(self) -> str:
        return f"{{}} {self.op.value} {{}}"


@dataclass
class Add(Arithmetic):
    op: ClassVar[ArithmeticOp] = ArithmeticOp.ADD


@dataclass
class Sub(Arithmetic):
    op: ClassVar[ArithmeticOp] = ArithmeticOp.SUB


@dataclass
class Mul(Arithmetic):
    op: ClassVar[ArithmeticOp] = ArithmeticOp.MUL


@dataclass
class Div(Arithmetic):
    op: ClassVar[ArithmeticOp] = ArithmeticOp.DIV


@dataclass
class Function(Expression):
    """ Node class to represent function invocations.
    """

    name: str
    arguments: List[AstType]

    def get_sub_nodes(self) -> List[AstType]:
        return self.arguments

    def get_template(self) -> str:
        return f"{self.name} ({', '.join(['{}'] * len(self.arguments))})"


def indent(text: str, amount: int, ch: str = ' ') -> str:
    padding = amount * ch
    return ''.join(padding + line for line in text.splitlines(True))


def get_repr(node: Node, indent_amount: int = 0, indent_incr: int = 4) -> str:
    """ Get a debug representation of the given AST node. ``indent_amount``
        and ``indent_incr`` are for the recursive call and don't need to be
        passed.
    """
    sub_nodes = node.get_sub_nodes()
    template = node.get_template()

    args = []
    for sub_node in sub_nodes:
        if isinstance(sub_node, Node) and not sub_node.inline:
            args.append(
                "(\n{}\n)".format(
                    indent(
                        get_repr(
                            sub_node,
                            indent_amount + indent_incr,
                            indent_incr
                        ),
                        indent_amount + indent_incr
                    )
                )
            )
        else:
            args.append(repr(sub_node))

    return template.format(*args)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>,
# David Bitner <bitner@dbspatial.com>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from typing import Dict, Optional
from datetime import datetime, date

from ..evaluator import Evaluator, handle
from ... import ast
from ...cql2 import get_op
from ... import values
import json


def json_serializer(obj):
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    if hasattr(obj, "name"):
        return obj.name
    raise TypeError(f"{obj} with type {type(obj)} is not serializable.")


class CQL2Evaluator(Evaluator):
    def __init__(
        self,
        attribute_map: Optional[Dict[str, str]],
        function_map: Optional[Dict[str, str]],
    ):
        self.attribute_map = attribute_map
        self.function_map = function_map

    @handle(
        ast.Condition,
        ast.Comparison,
        ast.TemporalPredicate,
        ast.SpatialComparisonPredicate,
        ast.Arithmetic,
        ast.ArrayPredicate,
        subclasses=True,
    )
    def comparison(self, node, *args):
        op = get_op(node)
        return {"op": op, "args": [*args]}

    @handle(ast.Between)
    def between(self, node, lhs, low, high):
        return {"op": "between", "args": [lhs, [low, high]]}

    @handle(ast.Like)
    def like(self, node, *subargs):
        return {"op": "like", "args": [node.lhs, node.pattern]}

    @handle(ast.IsNull)
    def isnull(self, node, arg):
        return {"op": "isNull", "args": arg}

    @handle(ast.Function)
    def function(self, node, *args):
        name = node.name.lower()
        if name == "lower":
            ret = {"lower": args[0]}
        else:
            ret = {"function": name, "args": [*args]}
        return ret

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        return {"in": {"value": lhs, "list": options}}

    @handle(ast.Attribute)
    def attribute(self, node: ast.Attribute):
        return {"property": node.name}

    @handle(values.Interval)
    def interval(self, node: values.Interval, start, end):
        return {"interval": [start, end]}

    @handle(datetime)
    def datetime(self, node: ast.Attribute):
        return {"timestamp": node.name}

    @handle(*values.LITERALS)
    def literal(self, node):
        return node

    @handle(values.Geometry)
    def geometry(self, node: values.Geometry):
        return node.__geo_interface__

    @handle(values.Envelope)
    def envelope(self, node: values.Envelope):
        return node.__geo_interface__


def to_cql2(
    root: ast.Node,
    field_mapping: Optional[Dict[str, str]] = None,
    function_map: Optional[Dict[str, str]] = None,
) -> str:
    return json.dumps(
        CQL2Evaluator(field_mapping, function_map).evaluate(root),
        default=json_serializer,
    )



from datetime import date, datetime, timedelta
from typing import Dict, Union
from lxml import etree

from ... import values
from ...util import parse_duration, parse_datetime
from .util import Element

Temporal = Union[date, datetime, timedelta, values.Interval]


def _parse_time_position(node: Element, nsmap: Dict[str, str]) -> datetime:
    return parse_datetime(node.text)


def _parse_time_instant(node: Element, nsmap: Dict[str, str]) -> datetime:
    position = node.xpath('gml:timePosition', namespaces=nsmap)[0]
    return _parse_time_position(position, nsmap)


def _parse_time_period(node: Element,
                       nsmap: Dict[str, str]) -> values.Interval:
    begin = node.xpath(
        'gml:begin/gml:TimeInstant/gml:timePosition|gml:beginPosition',
        namespaces=nsmap
    )[0]
    end = node.xpath(
        'gml:end/gml:TimeInstant/gml:timePosition|gml:endPosition',
        namespaces=nsmap
    )[0]
    return values.Interval(
        _parse_time_position(begin, nsmap),
        _parse_time_position(end, nsmap),
    )


def _parse_valid_time(node: Element, nsmap: Dict[str, str]) -> Temporal:
    return parse_temporal(node[0], nsmap)


def _parse_duration(node: Element, nsmap: Dict[str, str]) -> timedelta:
    return parse_duration(node.text)


PARSER_MAP = {
    'validTime': _parse_valid_time,
    'timePosition': _parse_time_position,
    'TimeInstant': _parse_time_instant,
    'TimePeriod': _parse_time_period,
    'duration': _parse_duration,
}


def is_temporal(node: Element) -> bool:
    return etree.QName(node).localname in PARSER_MAP


def parse_temporal(node: Element, nsmap: Dict[str, str]) -> Temporal:
    parser = PARSER_MAP[etree.QName(node).localname]
    return parser(node, nsmap)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from datetime import date, time, datetime, timedelta

from shapely import geometry

from . import filters
from ... import ast
from ... import values
from ..evaluator import Evaluator, handle


LITERALS = (str, float, int, bool, datetime, date, time, timedelta)


class GeoPandasEvaluator(Evaluator):
    def __init__(self, df, field_mapping=None, function_map=None):
        self.df = df
        self.field_mapping = field_mapping
        self.function_map = function_map

    @handle(ast.Not)
    def not_(self, node, sub):
        return filters.negate(sub)

    @handle(ast.And, ast.Or)
    def combination(self, node, lhs, rhs):
        return filters.combine((lhs, rhs), node.op.value)

    @handle(ast.Comparison, subclasses=True)
    def comparison(self, node, lhs, rhs):
        return filters.compare(
            lhs,
            rhs,
            node.op.value,
        )

    @handle(ast.Between)
    def between(self, node, lhs, low, high):
        return filters.between(
            lhs,
            low,
            high,
            node.not_
        )

    @handle(ast.Like)
    def like(self, node, lhs):
        return filters.like(
            lhs,
            node.pattern,
            node.nocase,
            node.wildcard,
            node.singlechar,
            node.escapechar,
            node.not_,
        )

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        return filters.contains(
            lhs, options, node.not_,
        )

    @handle(ast.IsNull)
    def null(self, node, lhs):
        return filters.null(
            lhs,
            node.not_,
        )

    @handle(ast.TemporalPredicate, subclasses=True)
    def temporal(self, node, lhs, rhs):
        return filters.temporal(
            node.lhs,
            node.rhs,
            node.op.value,
        )

    @handle(ast.SpatialComparisonPredicate, subclasses=True)
    def spatial_operation(self, node, lhs, rhs):
        return filters.spatial(
            lhs,
            rhs,
            node.op.name,
        )

    @handle(ast.BBox)
    def bbox(self, node, lhs):
        return filters.bbox(
            lhs,
            node.minx,
            node.miny,
            node.maxx,
            node.maxy,
            node.crs
        )

    @handle(ast.Attribute)
    def attribute(self, node):
        return filters.attribute(
            self.df,
            node.name,
            self.field_mapping
        )

    @handle(ast.Arithmetic, subclasses=True)
    def arithmetic(self, node, lhs, rhs):
        return filters.arithmetic(
            lhs,
            rhs,
            node.op.value
        )

    @handle(ast.Function)
    def function(self, node, *arguments):
        return self.function_map[node.name](*arguments)

    @handle(*values.LITERALS)
    def literal(self, node):
        return node

    @handle(values.Interval)
    def interval(self, node, start, end):
        return (start, end)

    @handle(values.Geometry)
    def geometry(self, node):
        return geometry.shape(node)

    @handle(values.Envelope)
    def envelope(self, node):
        return geometry.Polygon.from_bounds(
            node.x1, node.y1, node.x2, node.y2
        )


def to_filter(df, root, field_mapping=None, function_map=None):
    """
    """
    return GeoPandasEvaluator(df, field_mapping, function_map).evaluate(root)
